<?php
/**
 * AREST - RESTful Authentication for PHP 4/5
 * Version: 1.2.1
 * Author: Travis Estill (travisce.com)
 * Copyright (c) 2010 Travis Estill
 * License: Distributed under the Lesser General Public License
 * (LGPL) http://www.gnu.org/copyleft/lesser.html
 * This program is distributed in the hope that it will be useful
 * - WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 */

///////////////////////////////////////
// BUSINESS FUNCTIONS
///////////////////////////////////////

/**
 * Processes HTTP authentication
 * @param   string  $realm  authentication realm
 * @param   int     $flags  bitwise flags
 * @return  mixed           null or array (0=status, 1=message, 
 *                          2=nonce, 3=flags)
 */
function arest_auth($realm, $flags = 0) {
  global $arest_cfg,
         $arest_mysql;
  
  arest_set_error_handler();
  
  // Database connection
  $arest_mysql = mysql_connect(
    AREST_MYSQL_HOST, AREST_MYSQL_USER, AREST_MYSQL_PWD);
  if (!mysql_select_db(AREST_MYSQL_DB, $arest_mysql)) {
    trigger_error('Unable to select database', E_USER_WARNING);
  }
  
  // Set timezones to UTC
  mysql_query('SET time_zone = "+0:00"', $arest_mysql);
  define('AREST_TZ_RESET', getenv('TZ'));
  putenv('TZ=UTC');
  
  $arest_cfg['realm'] = $realm;
  arest_srand();
  
  // Cross-domain authentication
  if (isset($_GET['arest-user']) && isset($_GET['arest-nonce'])) {
    $nonce = rawurldecode($_GET['arest-nonce']);
    arest_view(403, '', $nonce, AREST_QUERY_AUTH);
    arest_cleanup();
    exit;
  }
  
  $result = arest_validate_auth_header($realm);
  $rstatus = $result[0]; // Result (r) status
  $msg = $result[1];
  $digest = $result[2];
  $auth = $result[3];
  $vflags = 0; // View (v) flags
  if (isset($auth['error_log'])) {
    arest_log_failure($digest, $auth['error_log']);
  }
  
  // Recognized Digest directives
  define('AREST_DIGEST_USERNAME',
    isset($digest['username']) ? $digest['username'] : '');
  define('AREST_DIGEST_REALM',
    isset($digest['realm']) ? $digest['realm'] : '');
  define('AREST_DIGEST_NONCE',
    isset($digest['nonce']) ? $digest['nonce'] : '');
  define('AREST_DIGEST_URI',
    isset($digest['uri']) ? $digest['uri'] : '');
  define('AREST_DIGEST_RESPONSE',
    isset($digest['response']) ? $digest['response'] : '');
  define('AREST_DIGEST_ALGORITHM',
    isset($digest['algorithm']) ? $digest['algorithm'] : '');
  define('AREST_DIGEST_CNONCE',
    isset($digest['cnonce']) ? $digest['cnonce'] : '');
  define('AREST_DIGEST_OPAQUE',
    isset($digest['opaque']) ? $digest['opaque'] : '');
  define('AREST_DIGEST_QOP',
    isset($digest['qop']) ? $digest['qop'] : '');
  define('AREST_DIGEST_NC',
    isset($digest['nc']) ? $digest['nc'] : '');
  
  // XHR authentication requires an initial 401 response
  if ($rstatus == 401 && empty($digest)) {
    if (isset($_POST['arest-auth-scheme'])) {
      if ($_POST['arest-auth-scheme'] == 'Basic') {
        $vflags = $vflags | AREST_BASIC_AUTH;
        $nonce = NULL;
      } else {
        $nonce = arest_create_nonce();
      }
      arest_view(401, $msg, $nonce, $vflags);
      arest_cleanup();
      exit;
    } else {
      $nonce = arest_create_nonce();
    }
  }
  
  if (!isset($nonce)) {
    if (isset($auth['nonce']) && $auth['nonce'] !== 1) {
      // Stale or invalid nonce sent
      $nonce = arest_create_nonce();
      $vflags = $vflags | AREST_RETRY_NONCE;
    } else if (!empty($digest['nonce'])) {
      // Use nonce from Authorization header
      $nonce = $digest['nonce'];
    } else {
      $nonce = NULL;
    }
  }
  
  $cname = 'arest-realm-' . md5($arest_cfg['realm']);
  
  if (isset($_POST['arest-challenge'])
        && $_POST['arest-challenge'] == TRUE) {
    $forbidden = 401;
    $vflags = $vflags | AREST_FORCE_401;
  } else if (isset($_COOKIE[$cname])
      && $_COOKIE[$cname] == 'Force-401') {
    $forbidden = 401;
  } else {
    $forbidden = 403;
  }
  
  if ($rstatus === 200) {
    $vstatus = 200;
    define('AREST_DIGEST_HA1', $auth['ha1']);
  } else if ($rstatus == 401) {
    $vstatus = $forbidden;
  } else {
    $vstatus = $result[0];
  }
  
  // Force 401 response for certain invalid Digest requests
  if ($vstatus == 403 && !empty($digest)) {
    if (empty($digest['qop']) || $digest['qop'] != 'auth-basic') {
      $vstatus = 401;
    }
  }
  
  if ($rstatus === 200) {
    if (isset($_POST['arest-disable-nonce'])) {
      // User logout
      $disable_hash = md5(
        $_POST['arest-disable-nonce'] . ':' . AREST_NONCE_SALT);
      arest_db_disable_nonce($disable_hash);
      $nonce = arest_create_nonce();
      $vstatus = 403;
      if (empty($digest['qop']) || $digest['qop'] != 'auth-basic') {
        $vflags = $vflags | AREST_UA_CACHE;
        $msg = $arest_cfg['msg_ua_cache'];
      } else {
        $msg = $arest_cfg['msg_nonce_disabled'];
      }
    } else {
      // Update nonce information
      if (AREST_NC_COMPAT == TRUE && isset($digest['nc'])) {
        $nc = hexdec($digest['nc']);
      } else {
        $nc = 0;
      }

      $expire = time() + AREST_NONCE_EXPIRE;
      $hash = md5($digest['nonce'] . ':' . AREST_NONCE_SALT);
      arest_db_update_nonce(
        $auth['user_id'], $expire, $nc, $hash);
    }
  }
  
  if (isset($_POST['arest-response'])
        && $_POST['arest-response'] == TRUE) {
    $return = FALSE;
  } else if ($vstatus === 200) {
    $return = TRUE;
  } else {
    $return = FALSE;
  }
  
  if ($return == FALSE) {
    arest_view($vstatus, $msg, $nonce, $vflags);
    arest_cleanup();
    exit;
  } else {
    return array($vstatus, $msg, $nonce, $vflags);
  }
}

/**
 * Sends the HTTP Authentication-Info header
 * @param   string       $entity  entity body of response
 * @return  string|bool           header content on success
 *                                or FALSE on failure
 */
function arest_auth_info($entity) {
  if (!defined('AREST_DIGEST_QOP')
        || AREST_DIGEST_QOP == 'auth-basic') {
    return FALSE;
  }
  
  $a2 = ":" . AREST_REQUEST_URI;
  if (AREST_DIGEST_QOP == 'auth-int') {
    $a2 .= ':' . md5($entity);
  }
  $ha2 = md5($a2);
  
  if (AREST_DIGEST_QOP == 'auth'
        || AREST_DIGEST_QOP == 'auth-int') {
    $rspauth = md5(AREST_DIGEST_HA1 . ':'
      . AREST_DIGEST_NONCE . ':'
      . AREST_DIGEST_NC . ':'
      . AREST_DIGEST_CNONCE . ':'
      . AREST_DIGEST_QOP . ':'
      . $ha2);
  } else if (AREST_DIGEST_QOP == '') {
    $rspauth = md5(AREST_DIGEST_HA1 . ':'
      . AREST_DIGEST_NONCE . ':'
      . $ha2);
  }
  
  if (AREST_DIGEST_QOP == 'auth'
        || AREST_DIGEST_QOP == 'auth-int') {
    return arest_view_auth_info(AREST_DIGEST_QOP, $rspauth,
      AREST_DIGEST_CNONCE, AREST_DIGEST_NC);
  } else if (AREST_DIGEST_QOP == '') {
    return arest_view_auth_info(AREST_DIGEST_QOP, $rspauth);
  } else {
    return FALSE;
  }
}

/**
 * Reset variables, configuration options, and free up memory
 * @return  null
 */
function arest_cleanup() {
  global $arest_mysql;
  
  if (AREST_TZ_RESET != '') {
    putenv('TZ=' . AREST_TZ_RESET);
  }
  
  if (AREST_ERROR_HANDLING == TRUE) {
    error_reporting(AREST_ERROR_REPORT_RESET);
    ini_set('display_errors', AREST_ERROR_DISPLAY_RESET);
    restore_error_handler();
  }
  
  if (is_resource($arest_mysql)) {
    mysql_close($arest_mysql);
  }
}

/**
 * Concatenate components of a generic URI
 * @param   array  $uri_parts  one-dimensional associative array of
 *                             URI parts (allowed keys: 'scheme', 
 *                             'userinfo', 'host', 'port', 'path', 
 *                             'query', and 'fragment')
 * @return  string             concatenated URI
 */
function arest_concat_uri($uri_parts) {
  $uri = '';
  
  if (isset($uri_parts['scheme'])) {
    $uri .= $uri_parts['scheme'] . ':';
  }
  
  if (isset($uri_parts['host'])) {
    $uri .= '//';
    
    if (isset($uri_parts['userinfo'])) {
      $uri .= $uri_parts['userinfo'] . '@';
    }
    
    $uri .= $uri_parts['host'];
  
    if (isset($uri_parts['port'])) {
      $port = (int) $uri_parts['port'];
      $uri .= ':' . $port;
    }
  } else if (isset($uri_parts['authority'])) {
    $uri .= '//' . $uri_parts['authority'];
  }
  
  $uri .= (isset($uri_parts['path'])) ? $uri_parts['path'] : '';
  
  if (isset($uri_parts['query'])) {
    $uri .= '?' . $uri_parts['query'];
  }
  
  if (isset($uri_parts['fragment'])) {
    $uri .= '#' . $uri_parts['fragment'];
  }
  
  return $uri;
}

/**
 * Returns the response content type based on the HTTP Accept header
 * @return  string  content type
 */
function arest_content_type() {
  $accept = arest_parse_accept_header();
  usort($accept, 'arest_sort_accept_header');
  $num = count($accept);
  for ($i = 0; $i < $num; $i++) {
    $type = $accept[$i][1];
    $subtype = $accept[$i][2];
    if ($type == 'text' && $subtype == 'html') {
      return 'text/html';
    } else if ($type == 'text' && $subtype == 'plain') {
      return 'text/plain';
    }
  }
  
  return 'text/html';
}

/**
 * Creates new, unique nonce and adds it to the database
 * Raises an E_USER_WARNING on failure
 * @return  string|bool  nonce value on success or FALSE on failure
 */
function arest_create_nonce() {
  // Infrequently run maintenance query
  $odds = ceil(1 / AREST_NONCE_MAINT_ODDS);
  if (mt_rand(1, $odds) == 1) {
    arest_db_delete_old_nonces();
  }
  
  $nonce = arest_rand_str(16);
  $hash = md5($nonce . ':' . AREST_NONCE_SALT);
  
  $result = arest_db_add_nonce($hash, NULL,
              time() + AREST_NONCE_EXPIRE);
  if ($result == TRUE) {
    return $nonce;
  } else {
    trigger_error(
      'Could not add new nonce to database', E_USER_WARNING);
    return FALSE;
  }
}

/**
 * Default domain for Digest response
 * @return  string  default domain
 */
function arest_default_domain() {
  $uri = arest_parse_uri(AREST_REQUEST_URI);
  
  // Absolute and relative URIs
  $abs = $rel = $uri;
  
  if (AREST_SERVER_PORT == AREST_HTTPS_PORT) {
    $abs['scheme'] = 'https';
  } else {
    $abs['scheme'] = 'http';
  }
  
  if (substr(AREST_HTTP_HOST, 0, 4) == 'www.') {
    $abs['host'] = substr(AREST_HTTP_HOST, 4);
  } else {
    $abs['host'] = 'www.' . AREST_HTTP_HOST;
  }
  
  $abs['port'] = AREST_SERVER_PORT;
  
  $domain = array(arest_concat_uri($abs), arest_concat_uri($rel));
  
  return $domain;
}

/**
 * Handles server errors
 * @param   int          $error_num   error level
 * @param   string       $error_str   error message
 * @param   string|null  $error_file  (optional) path to file where 
 *                                    error occurred
 * @param   int|null     $error_line  (optional) line number the 
 *                                    error occurred at
 * @return  null
 */
function arest_error_handler($errno, $errstr, $errfile, $errline) {
  $error_msg = $errstr;
  if (!is_null($errfile) && !is_null($errline)) {
    $error_msg .= ' in ' . $errfile;
    $error_msg .= ' on line ' . $errline;
  }
  $code = substr(md5($error_msg), 0, 8);
  $debug = $error_msg;
  
  if (AREST_ERROR_LOG !== '') {
    $error_log = date('Y-m-d H:i:s \U\T\C') . ' - Server Error: ';
    $error_log .= $error_msg;
    $error_log .= ' (code:' . $code . ')';
    $error_log = wordwrap($error_log, 70, "\n");
    
    $log_file = AREST_ERROR_LOG . '/' . date('Y-m-d') . '.log';
    if (is_writable(AREST_ERROR_LOG)) {
      if (error_log($error_log . "\n", 3, $log_file)) {
        $debug = "Error Code: {$code}";
      }
    }
  }
  
  arest_ob_end_clean_all();
  
  $content_type = arest_content_type();
  
  if (!headers_sent()) {
    header("HTTP/1.1 500 Internal Server Error");
    header("Content-Type: {$content_type}; charset=iso-8859-1");
    header("Vary: Accept");
  }
  
  $message = <<<TXT
The authentication service failed to complete your request.
Please contact the server administrator, and provide the time 
and date of this occurrence, in addition to any further
diagnostic information.
TXT;
  
  if ($content_type == 'text/plain') {
    print "500 Internal Server Error\n\n";
    print $message . "\n\n";
    print $debug;
  } else if ($content_type == 'text/html') {
    print <<<HTML
<!DOCTYPE HTML PUBLIC "-//IETF//DTD HTML 2.0//EN">
<html><head>
<title>500 Internal Server Error</title>
</head><body>
<h1>Internal Server Error</h1>
<p>{$message}</p>
<p>{$debug}</p>
</body></html>
HTML;
  }
  
  arest_cleanup();
  exit;
}

/**
 * Converts HTTP status code into its reason phrase
 * @param   int     $code  status code
 * @return  string         reason phrase
 */
function arest_http_status_phrase($code) {
  $phrases = array(
    100 => 'Continue',
    101 => 'Switching Protocols',
    200 => 'OK',
    201 => 'Created',
    202 => 'Accepted',
    203 => 'Non-Authoritative Information',
    204 => 'No Content',
    205 => 'Reset Content',
    206 => 'Partial Content',
    300 => 'Multiple Choices',
    301 => 'Moved Permanently',
    302 => 'Found',
    303 => 'See Other',
    304 => 'Not Modified',
    305 => 'Use Proxy',
    307 => 'Temporary Redirect',
    400 => 'Bad Request',
    401 => 'Unauthorized',
    402 => 'Payment Required',
    403 => 'Forbidden',
    404 => 'Not Found',
    405 => 'Method Not Allowed',
    406 => 'Not Acceptable',
    407 => 'Proxy Authentication Required',
    408 => 'Request Time-out',
    409 => 'Conflict',
    410 => 'Gone',
    411 => 'Length Required',
    412 => 'Precondition Failed',
    413 => 'Request Entity Too Large',
    414 => 'Request-URI Too Large',
    415 => 'Unsupported Media Type',
    416 => 'Requested range not satisfiable',
    417 => 'Expectation Failed',
    500 => 'Internal Server Error',
    501 => 'Not Implemented',
    502 => 'Bad Gateway',
    503 => 'Service Unavailable',
    504 => 'Gateway Time-out',
    505 => 'HTTP Version not supported');
  
  if (isset($phrases[$code])) {
    return $phrases[$code];
  } else {
    return 'Unknown Status';
  }
}

/**
 * Logs authentication failure
 * @param   array|null  $digest  authentication data (optional)
 * @param   string      $reason  reason for failure (optional)
 * @return  bool                 TRUE on success or FALSE on failure
 */
function arest_log_failure($digest = NULL, $reason = '') {
  $username = isset($digest['username']) ? $digest['username'] : '';
  $response = isset($digest['response']) ? $digest['response'] : '';
  
  if (AREST_ERROR_LOG !== '') {
    $error = date('Y-m-d H:i:s \U\T\C') . ' - Login Failure: ';
    $error .= 'reason="' . $reason . '", ';
    $error .= 'URI=' . AREST_REQUEST_URI . ', ';
    $error .= 'IPaddr=' . AREST_REMOTE_ADDR . ', ';
    $error .= 'username="' . $username . '", ';
    $error .= 'response=' . $response;
    $error = wordwrap($error, 70, "\n");
    $log_file = AREST_ERROR_LOG . '/' . date('Y-m-d') . '.log';
    if (is_writable(AREST_ERROR_LOG)) {
      return error_log($error . "\n", 3, $log_file);
    }
  } else {
    return FALSE;
  }
}

/**
 * Erases and disables all output buffering
 * @return  null
 */
function arest_ob_end_clean_all() {
  $handlers = ob_list_handlers();
  $num = count($handlers);
  for ($i = 0; $i < $num; $i++) {
    ob_end_clean();
  }
}

/**
 * Parses the HTTP Authorization header
 * @return  mixed  authentication details (array) on success,
 *                 NULL on missing Authorization header, 
 *                 FALSE on parse failure
 */
function arest_parse_auth_header() {
  if (!empty($_SERVER['HTTP_AUTHORIZATION'])) {
    $auth_parts = preg_split(
      '/\s+/', $_SERVER['HTTP_AUTHORIZATION'], 2);
    if (strtolower($auth_parts[0]) == 'basic') {
      // Basic authentication
      $decode = base64_decode($auth_parts[1]);
      $credentials = explode(':', $decode, 2);
      if (count($credentials) == 2) {
        $_SERVER['PHP_AUTH_USER'] = $credentials[0];
        $_SERVER['PHP_AUTH_PW'] = $credentials[1];
      } else {
        return FALSE;
      }
    } else if (strtolower($auth_parts[0]) == 'digest') {
      // Digest authentication
      $_SERVER['PHP_AUTH_DIGEST'] = $auth_parts[1];
    }
  }
  
  if (isset($_SERVER['PHP_AUTH_USER'])
        && isset($_SERVER['PHP_AUTH_PW'])) {
    $digest = base64_decode($_SERVER['PHP_AUTH_PW']);
  } else if (isset($_SERVER['PHP_AUTH_DIGEST'])) {
    $digest = $_SERVER['PHP_AUTH_DIGEST'];
  } else {
    return NULL;
  }
  
  if ($digest === FALSE) {
    return FALSE;
  }
  
  // HTTP "token" (rfc2616 sec2.2)
  $token = 'A-Za-z0-9';
  $token .= '\x23-\x27';
  $token .= '\x5E-\x60';
  $token .= '\x21\x2A\x2B\x2D\x2E\x7C\x7E';
  
  $regexp = '@([' . $token . ']+)=';
  $regexp .= '(?:(?:"([^"\\\\]*(?:\\\\.[^"\\\\]*)*)")';
  $regexp .= '|([' . $token . ']+))@';
  
  preg_match_all($regexp, $digest, $matches, PREG_SET_ORDER);
  
  $digest_parts = array();
  if (count($matches) > 0) {
    foreach ($matches as $m) {
      $part = (isset($m[3])) ? $m[3] : $m[2];
      $part = preg_replace('/\\\\([^$])/', '$1', $part);
      $digest_parts[$m[1]] = $part;
    }
  } else {
    return FALSE;
  }
  
  return $digest_parts;
}

/**
 * Parses the HTTP Accept header
 * @return  array  media types
 *                 array([0..n] =>
 *                   array(
 *                     [0] => content-type,
 *                     [1] => type,
 *                     [2] => subtype,
 *                     [3] => position,
 *                     [4] => q,
 *                     [5..n] => extensions))
 */
function arest_parse_accept_header() {
  // Set default output array
  $accept = array();
  $accept[0] = array(0 => '*/*', 1 => '*', 2 => '*', 3 => 0, 4 => 1);
  
  if (empty($_SERVER['HTTP_ACCEPT'])) {
    return $accept;
  }
  
  // Split by commas
  $ranges = preg_split('/(,)(?=(?:[^"]|"[^"]*")*$)/m', 
    $_SERVER['HTTP_ACCEPT']);
  $rnum = count($ranges);
  for ($i = 0; $i < $rnum; $i++) {
    $accept[$i] = array(0 => trim($ranges[$i])); // Full media-range
    
    // Split by semi-colons
    $params = preg_split('/(;)(?=(?:[^"]|"[^"]*")*$)/m',
      $ranges[$i]);
    
    // Parse content-type
    $content_type = trim($params[0]);
    $types = explode('/', $content_type);
    $accept[$i][1] = (isset($types[0]) && strlen($types[0]) > 0)
                     ? $types[0]
                     : '*';
    $accept[$i][2] = (isset($types[1]) && strlen($types[1]) > 0)
                     ? $types[1]
                     : '*';
    
    array_shift($params); // Omit content-type (first element)
    $accept[$i][3] = $i; // Position (index)
    $accept[$i][4] = 1; // Default quality (q)
    unset($quality);
    $pnum = count($params);
    for ($j = 0; $j < $pnum; $j++) {
      // Match quality attribute, or append attribute to output array
      if (!isset($quality)) {
        $regexp = '/^\s*q=([0-1](\.[0-9]{0,3})?)/';
        $result = preg_match($regexp, $params[$j], $q);
      } else {
        $result = 0;
      }
      if ($result === 1) {
        // Truncate to 3 decimal places (x.xxx)
        $quality = (float) substr($q[1], 0, 5);
        if ($quality > 1) {
          $quality = 1;
        }
        $accept[$i][4] = $quality;
      } else {
        $accept[$i][] = trim($params[$j]);
      }
    }
  }
  
  return $accept;
}

/**
 * Parses generic URI and returns its components
 * @param   string  $uri        URI
 * @param   string  $component  if not empty, function will return 
 *                              the given component
 * @return  mixed               one-dimensional associative array of 
 *                              URI components OR single 
 *                              component (string)
 */
function arest_parse_uri($uri, $component = '') {
  $regexp = '|^(([^:/?#]+):)?'    // Scheme
              . '(//('            // Authority (BEGIN)
                . '(([^/?#]*)@)?'   // User Information
                . '([^:/?#]*)'      // Host
                . '(:([^/?#]*))?'   // Port
              . '))?'             // Authority (END)
              . '([^?#]*)'        // Path
              . '(\?([^#]*))?'    // Query
              . '(#(.*))?|';      // Fragment
  
  preg_match($regexp, $uri, $matches);
  
  $components = array();
  
  if (!empty($matches[2])) {
    $components['scheme'] = $matches[2];
  }
  
  if (!empty($matches[4])) {
    $components['authority'] = $matches[4];
  }
  
  if (!empty($matches[6])) {
    $components['userinfo'] = $matches[6];
  }
  
  if (!empty($matches[7])) {
    $components['host'] = $matches[7];
  }
  
  if (!empty($matches[9])) {
    $components['port'] = $matches[9];
  }
  
  if (!empty($matches[10])) {
    $components['path'] = $matches[10];
  }
  
  if (!empty($matches[12])) {
    $components['query'] = $matches[12];
  }
  
  if (!empty($matches[14])) {
    $components['fragment'] = $matches[14];
  }
  
  if (!empty($component)) {
    return $components[$component];
  } else {
    return $components;
  }
}

/**
 * Generates random string
 * @param   int     $rand_len  length of random string
 * @return  string             random string
 */
function arest_rand_str($rand_len = 16) {
  $options = '0123456789-._';
  $options .= 'ABCDEFGHIJKLMNOPQRSTUVWXYZ';
  $options .= 'abcdefghijklmnopqrstuvwxyz';
  $options_len = strlen($options);
  
  $rand_str = '';
  
  for ($i = 0; $i < $rand_len; $i++) {
    $rand_str .= substr(
      $options, mt_rand(0, ($options_len - 1)), 1);
  }
  
  return $rand_str;
}

/**
 * Returns the entity body from the HTTP request
 * @return  string  entity body
 */
function arest_request_entity_body() {
  return @file_get_contents('php://input');
}

/**
 * Configures the error handler
 * @return  null
 */
function arest_set_error_handler() {
  if (AREST_ERROR_HANDLING == TRUE) {
    $reset = error_reporting(E_ALL);
    if (!defined('AREST_ERROR_REPORT_RESET')) {
      define('AREST_ERROR_REPORT_RESET', $reset);
    }
    $reset = ini_set('display_errors', 0);
    if (!defined('AREST_ERROR_DISPLAY_RESET')) {
      define('AREST_ERROR_DISPLAY_RESET', $reset);
    }
    set_error_handler('arest_error_handler');
  }
}

/**
 * Sorts the parsed HTTP Accept header
 * Usage: usort($accept, "arest_sort_accept_header")
 * @param   array  $a  first media-range
 * @param   array  $b  second media-range
 * @return  int        (-1, 0, 1) according to "usort()"
 */
function arest_sort_accept_header($a, $b) {
  // Compare quality (q) values; descending
  if ($a[4] > $b[4]) {
    return -1;
  } else if ($a[4] < $b[4]) {
    return 1;
  }
  
  // Compare extension specificity; descending
  $anum = count($a);
  $bnum = count($b);
  if ($anum > $bnum) {
    return -1;
  } else if ($anum < $bnum) {
    return 1;
  }
  
  // Compare asterisks in type; descending
  if ($a[1] != '*' && $b[1] == '*') {
    return -1;
  } else if ($a[1] == '*' && $b[1] != '*') {
    return 1;
  }
  
  // Compare asterisks in subtype; descending
  if ($a[2] != '*' && $b[2] == '*') {
    return -1;
  } else if ($a[2] == '*' && $b[2] != '*') {
    return 1;
  }
      
  // Compare positions; ascending
  if ($a[3] > $b[3]) {
    return 1;
  } else if ($a[3] < $b[3]) {
    return -1;
  }
  
  return 0;
}

/**
 * Seeds the random number generator
 * @return  null
 */
function arest_srand() {
  if (function_exists('microtime')) {
    list($usec, $sec) = explode(' ', microtime());
    $seed = (float) $sec + ((float) $usec * 100000);
    if (function_exists('posix_getpid')) {
      $seed += (float) posix_getpid();
    }
    mt_srand($seed);
  }
}

/**
 * Validates HTTP Authorization header
 * @param   string  $realm  authentication realm
 * @param   int     $flags  bitwise flags
 * @return  array           0=status, 1=message, 2=digest-params, 
 *                          3=auth-data
 */
function arest_validate_auth_header($realm, $flags = 0) {
  global $arest_cfg,
         $arest_mysql;
  
  $auth = array();
  
  // Check for Authorization header
  $digest = arest_parse_auth_header();
  if (is_null($digest)) {
    return array(401, $arest_cfg['msg_auth_required'], 
      array(), $auth);
  } else if ($digest === FALSE) {
    return array(400, $arest_cfg['msg_auth_malformed'], 
      array(), $auth);
  }
  
  $digest['qop'] = (isset($digest['qop'])) ? $digest['qop'] : '';
  if (AREST_AUTH_BASIC_COMPAT == FALSE
        && $digest['qop'] == 'auth-basic') {
    return array(501, $arest_cfg['msg_auth_basic_incompat'], 
      $digest, $auth);
  }
  if (AREST_AUTH_INT_COMPAT == FALSE
        && $digest['qop'] == 'auth-int') {
    return array(501, $arest_cfg['msg_auth_int_incompat'], 
      $digest, $auth);
  }
  
  // Check for missing Digest directives
  switch ($digest['qop']) {
    case 'auth-basic':
      $req = array('username', 'realm', 'nonce', 'response',
        'cnonce');
      break;
    case 'auth':
    case 'auth-int':
      $req = array('username', 'realm', 'nonce', 'uri', 'response',
        'cnonce', 'nc');
      break;
    case '':
      $req = array('username', 'realm', 'nonce', 'uri', 'response');
      break;
    default:
      return array(400, $arest_cfg['msg_auth_malformed'], 
        $digest, $auth);
      break;
  }
  
  // IE6 does not send required "opaque" directive
  if (AREST_IE6_COMPAT == FALSE) {
    array_push($req, 'opaque');
  }
  
  $num = count($req);
  for ($i = 0; $i < $num; $i++) {
    if (!isset($digest[$req[$i]])) {
      return array(400, $arest_cfg['msg_auth_malformed'], 
        $digest, $auth);
    }
  }
  
  // Validate Digest directives
  $user_id = arest_db_get_user_id($digest['username']);
  $auth['user_id'] = $user_id;
  if ($user_id == FALSE) {
    return array(401, $arest_cfg['msg_auth_invalid'], 
      $digest, $auth);
  }
  
  if ($digest['realm'] != $realm) {
    return array(401, $arest_cfg['msg_realm_invalid'], 
      $digest, $auth);
  }
  
  if ($digest['qop'] != 'auth-basic') {
    // IE6 does not include query string in "digest-uri"
    if (AREST_IE6_COMPAT == TRUE) {
      $digest_path = arest_parse_uri($digest['uri'], 'path');
      $request_path = arest_parse_uri(AREST_REQUEST_URI, 'path');
      if ($digest_path != $request_path) {
        $auth['error_log'] = 'URI mismatch';
        return array(400, $arest_cfg['msg_uri_invalid'], 
          $digest, $auth);
      }
    } else if ($digest['uri'] != AREST_REQUEST_URI) {
      return array(400, $arest_cfg['msg_uri_invalid'], 
        $digest, $auth);
    }
  }
  
  // IE6 does not send required "opaque" directive
  if (AREST_IE6_COMPAT == FALSE
        && $digest['opaque'] != AREST_OPAQUE) {
    return array(401, $arest_cfg['msg_opaque_invalid'], 
      $digest, $auth);
  }
  
  // Validate Request-Digest
  $ha1 = arest_db_get_digest_ha1($user_id, $digest['realm']);
  $auth['ha1'] = $ha1;
  if ($digest['qop'] == 'auth' || $digest['qop'] == '') {
    $ha2 = md5(AREST_REQUEST_METHOD . ':' . $digest['uri']);
  } else if ($digest['qop'] == 'auth-int') {
    $body = md5(arest_request_entity_body());
    $ha2 = md5(AREST_REQUEST_METHOD . ':' . $digest['uri']
      . ':' . $body);
  }
  
  if ($digest['qop'] == 'auth-basic') {
    $valid = md5($ha1 . ':' . $digest['nonce']
      . ':' . $digest['cnonce']);
  } else if ($digest['qop'] == 'auth'
               || $digest['qop'] == 'auth-int') {
    $valid = md5($ha1 . ':' . $digest['nonce'] . ':' . $digest['nc']
      . ':' . $digest['cnonce'] . ':' . $digest['qop'] . ':' . $ha2);
  } else if ($digest['qop'] == '') {
    $valid = md5($ha1 . ':' . $digest['nonce'] . ':' . $ha2);
  }
  if ($valid != $digest['response']) {
    $auth['error_log'] = 'Invalid Request-Digest';
    return array(401, $arest_cfg['msg_auth_invalid'], 
      $digest, $auth);
  }
  
  // Validate nonce
  $hash = md5($digest['nonce'] . ':' . AREST_NONCE_SALT);
  $db_nonce = arest_db_get_nonce($hash);
  if (!$db_nonce) {
    $auth['nonce'] = -1; // Invalid nonce
    return array(401, $arest_cfg['msg_nonce_invalid'], 
      $digest, $auth);
  } else {
    if (!is_null($db_nonce['user_id'])
          && $db_nonce['user_id'] != $user_id) {
      $auth['nonce'] = -1;
      return array(401, $arest_cfg['msg_nonce_invalid'], 
        $digest, $auth);
    }
    
    if ($db_nonce['expiration'] < time()) {
      $auth['nonce'] = 0; // Stale nonce
      return array(401, $arest_cfg['msg_nonce_expired'], 
        $digest, $auth);
    }
    
    if ($db_nonce['disabled'] == TRUE) {
      $auth['nonce'] = 0;
      return array(401, $arest_cfg['msg_nonce_disabled'], 
        $digest, $auth);
    }
    
    $auth['nonce'] = 1; // Validated nonce
    
    if (AREST_NC_COMPAT == TRUE && isset($digest['nc'])) {
      if (hexdec($digest['nc']) <= $db_nonce['nc']) {
        return array(401, $arest_cfg['msg_nc_invalid'], 
          $digest, $auth);
      }
    }
  }
  
  return array(200, $arest_cfg['msg_auth_accepted'], 
    $digest, $auth);
}

if (!function_exists('http_build_query')) {
  function http_build_query(
             $data, $prefix = '', $sep = '', $key = '') {
    $ret = array();
    foreach ((array) $data as $k => $v) {
      if (is_int($k) && $prefix != NULL) {
        $k = urlencode($prefix . $k);
      }
      
      if ((!empty($key)) || ($key === 0)) {
        $k = $key . '[' .urlencode($k) . ']';
      }
      
      if (is_array($v) || is_object($v)) {
        array_push($ret, http_build_query($v, '', $sep, $k));
      } else {
        array_push($ret, $k . '=' . urlencode($v));
      }
    }
    
    if (empty($sep)) {
      $sep = ini_get('arg_separator.output');
    }
    
    return implode($sep, $ret);
  }
}
?>
